from typing import TYPE_CHECKING, Optional, Type

from .ast_node import ASTNode
from .conditional import is_castable

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

# Logical boolean operators:
#   && ||
# Both operands are self-determined
class _BoolExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', l: ASTNode, r: ASTNode):
        super().__init__(env, src_ref)
        self.l = l
        self.r = r

    def predict_type(self) -> Type[bool]:
        l_type = self.l.predict_type()
        r_type = self.r.predict_type()
        if not is_castable(l_type, bool):
            self.msg.fatal(
                "Left operand of expression is not a compatible boolean type",
                self.src_ref
            )
        if not is_castable(r_type, bool):
            self.msg.fatal(
                "Right operand of expression is not a compatible boolean type",
                self.src_ref
            )
        return bool

    def get_min_eval_width(self) -> int:
        return 1

class BoolAnd(_BoolExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l = bool(self.l.get_value())
        r = bool(self.r.get_value())
        return l and r

class BoolOr(_BoolExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l = bool(self.l.get_value())
        r = bool(self.r.get_value())
        return l or r



from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Any, Dict


if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from ..rdltypes.typing import PreElabRDLType

    OptionalSourceRef = Optional[SourceRefBase]

class ASTNode:
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef'):
        self.env = env
        self.msg = env.msg

        # Source Ref to use for error context
        self.src_ref = src_ref

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'ASTNode':
        """
        Deepcopy all members except for ones that should be copied by reference
        """
        copy_by_ref = ["env", "msg"]
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in copy_by_ref:
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def predict_type(self) -> 'PreElabRDLType':
        """
        Returns the expected type of the result of the expression
        This function shall call any child expression's predict_type()
        even if the child type is not relevant to the resulting type.
        Raises exception if input types are not compatible.
        """
        raise NotImplementedError

    def get_min_eval_width(self) -> int:
        """
        Returns the expressions resulting integer width based on the
        self-determined expression bit-width rules
        (SystemVerilog LRM: IEEE Std 1800-2012, Table 11-21)
        """
        raise RuntimeError

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        """
        Compute the value of the result of the expression

        Since predict_type() was already called, safe to assume that types are
        compatible.

        The eval_width parameter controls the integer width resolution context
        - If eval_width is irrelevant:
            For example, comparison operator, integer literal, non-integer expression)
            OK to ignore eval_width.
            Otherwise...
        - If eval_width is None:
            This expression node is the start of a new self-determined context.
            Query the relevant operands to determine the context's eval_width
        - If eval_width is set to a value:
            Parent expression is propagating the eval_width
        """
        raise NotImplementedError




from typing import Any, Optional, TYPE_CHECKING

from .builtin_enums import BuiltinEnum, AccessType, OnReadType, OnWriteType
from .builtin_enums import AddressingType, PrecedenceType, InterruptType

from .user_enum import UserEnumMemberContainer, UserEnum, is_user_enum

from .user_struct import UserStruct, is_user_struct

from .array import ArrayPlaceholder

from .references import ComponentRef, PropertyReference

if TYPE_CHECKING:
    from .typing import PreElabRDLType

#-------------------------------------------------------------------------------
def get_rdltype(value: Any) -> 'PreElabRDLType':
    """
    Given a value, return the type identifier object used within the RDL compiler
    If not a supported type, return None
    """

    if isinstance(value, (int, bool, str)):
        # Pass canonical types as-is
        return type(value)
    elif is_user_enum(type(value)):
        return type(value)
    elif is_user_struct(type(value)):
        return type(value)
    elif isinstance(value, BuiltinEnum):
        return type(value)
    elif isinstance(value, list):
        # Create ArrayPlaceholder representation
        # Determine element type and make sure it is uniform
        array_el_type = None # type: Optional[PreElabRDLType]
        for el in value:
            el_type = get_rdltype(el)
            if el_type is None:
                return None

            if (array_el_type is not None) and (el_type != array_el_type):
                return None
            array_el_type = el_type
        return ArrayPlaceholder(array_el_type)
    else:
        return None


class NoValue:
    """
    Non-value token used for UDPs that were bound to a component, but not
    actually assigned a value.
    (15.2.1-c)
    """



from typing import TYPE_CHECKING, Optional, Type

from .ast_node import ASTNode
from .conditional import is_castable

from ..core.helpers import truncate_int

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

# Reduction operators:
#   &  ~&  |  ~|  ^  ^~  !
# Result is always 1 bit int
# Creates a new evaluation context
class _ReductionExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', n: ASTNode):
        super().__init__(env, src_ref)
        self.n = n

    def predict_type(self) -> Type[int]:
        op_type = self.n.predict_type()
        if not is_castable(op_type, int):
            self.msg.fatal(
                "Operand of expression is not a compatible numeric type",
                self.src_ref
            )
        return int

    def get_min_eval_width(self) -> int:
        return 1

class AndReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        eval_width = self.n.get_min_eval_width()
        n = int(self.n.get_value(eval_width))
        n = truncate_int(~n, eval_width)
        return int(n == 0)

class NandReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        eval_width = self.n.get_min_eval_width()
        n = int(self.n.get_value(eval_width))
        n = truncate_int(~n, eval_width)
        return int(n != 0)

class OrReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        n = int(self.n.get_value())
        return int(n != 0)

class NorReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        n = int(self.n.get_value())
        return int(n == 0)

class XorReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        n = int(self.n.get_value())
        v = 0
        while n:
            if n & 1:
                v ^= 1
            n >>= 1
        return v

class XnorReduce(_ReductionExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        n = int(self.n.get_value())
        v = 1
        while n:
            if n & 1:
                v ^= 1
            n >>= 1
        return v

class BoolNot(_ReductionExpr):
    def predict_type(self) -> Type[bool]:
        super().predict_type()
        return bool

    def get_value(self, eval_width: Optional[int]=None) -> bool:
        n = int(self.n.get_value())
        return not n



from typing import TYPE_CHECKING, Type, Optional, Any, Tuple

from .ast_node import ASTNode
from .conditional import is_castable

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

#-------------------------------------------------------------------------------
# Relational operators:
#   == != < > <= >=
# Result is always 1 bit bool
# Creates a new evaluation context
# Child operands are evaluated in the same width context, sized to the max
# of either op.
class _RelationalExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', l: ASTNode, r: ASTNode):
        super().__init__(env, src_ref)
        self.l = l
        self.r = r
        self.is_numeric = None # type: bool

    def predict_type(self) -> Type[bool]:
        l_type = self.l.predict_type()
        r_type = self.r.predict_type()

        # Type of L and R operands shall be compatible
        if is_castable(l_type, int) and is_castable(r_type, int):
            self.is_numeric = True
        elif l_type == r_type:
            # Same types. Inherently compatible
            self.is_numeric = False
        else:
            # Incompatible
            self.msg.fatal(
                "Left and right operands of expression are not compatible types",
                self.src_ref
            )
        return bool

    def get_min_eval_width(self) -> int:
        return 1

    def get_ops(self) -> Tuple[Any, Any]:

        if self.is_numeric:
            # New width context. Determine eval_width here
            eval_width = max(
                self.l.get_min_eval_width(),
                self.r.get_min_eval_width()
            )

            l = int(self.l.get_value(eval_width))
            r = int(self.r.get_value(eval_width))
        elif not self.is_numeric:
            l = self.l.get_value()
            r = self.r.get_value()
        else:
            raise RuntimeError

        return l, r

class _NumericRelationalExpr(_RelationalExpr):

    def predict_type(self) -> Type[bool]:
        l_type = self.l.predict_type()
        r_type = self.r.predict_type()

        # Type of L and R operands shall be integral types
        self.is_numeric = True
        if not is_castable(l_type, int):
            self.msg.fatal(
                "Left operand of expression is not a compatible numeric type",
                self.src_ref
            )
        if not is_castable(r_type, int):
            self.msg.fatal(
                "Right operand of expression is not a compatible numeric type",
                self.src_ref
            )
        return bool



class Eq(_RelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l == r

class Neq(_RelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l != r

class Lt(_NumericRelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l < r

class Gt(_NumericRelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l > r

class Leq(_NumericRelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l <= r

class Geq(_NumericRelationalExpr):
    def get_value(self, eval_width: Optional[int]=None) -> bool:
        l, r = self.get_ops()
        return l >= r



from typing import Type, Union, Any, TYPE_CHECKING
import sys

if TYPE_CHECKING:
    from .builtin_enums import BuiltinEnum
    from .user_enum import UserEnum
    from .references import PropertyReference, RefType
    from .user_struct import UserStruct
    from .array import ArrayPlaceholder
    from ..node import Node
    from .. import component as comp

if sys.version_info >= (3,5,4):
    # RDL Types the user will encounter via the public API
    RDLValue = Union[
        int, bool, str, list,
        'BuiltinEnum', 'UserEnum', 'UserStruct', Type['UserEnum'],
        'Node', 'PropertyReference'
    ]
else:
    # Stub on 3.5.3 or older due to: https://github.com/python/typing/issues/266
    RDLValue = Any # type: ignore

# RDL Types used internally prior to elaboration
PreElabRDLValue = Union[
    int, bool, str,
    'BuiltinEnum', 'UserEnum', 'UserStruct',
    'PropertyReference', 'comp.Component', 'RefType'
]

if sys.version_info >= (3,5,4):
    PreElabRDLType = Union[Type[PreElabRDLValue], 'ArrayPlaceholder']
else:
    # Stub on 3.5.3 or older due to: https://github.com/python/typing/issues/266
    PreElabRDLType = Any # type: ignore



from typing import Tuple, Union, Dict, Any

from antlr4.Token import CommonToken
from antlr4 import ParserRuleContext
from antlr4.tree.Tree import TerminalNodeImpl

from .preprocessor.stream import PreprocessedInputStream
from .preprocessor.segment_map import SegmentMap

#-------------------------------------------------------------------------------
# Generic Base Classes
#-------------------------------------------------------------------------------
class SourceRefBase:
    """
    Base class for all source references
    """

    def derive_coordinates(self) -> None:
        """
        DEPRECATED
        Some dependents may call this if using the old SourceRef API
        """

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'SourceRefBase':
        # Don't deepcopy source refs
        return self


class FileSourceRef(SourceRefBase):
    """
    Simple file source reference that is only capable of providing the path to
    the originating file.

    Some register model importers may only be able to provide limited source context.
    """
    def __init__(self, path: str):
        super().__init__()
        self._path = path

    @property
    def path(self) -> str:
        """
        Path of the originating file
        """
        return self._path

    @property
    def filename(self) -> str:
        """
        DEPRECATED
        Alias to ``path``.
        Some dependents may call this if using the old SourceRef API
        """
        return self.path

class DetailedFileSourceRef(FileSourceRef):
    """
    Detailed source reference that is capable of providing a context snippet
    from the originating source.
    """

    @property
    def path(self) -> str:
        """
        Path of the originating file
        """
        raise NotImplementedError

    @property
    def line(self) -> int:
        """
        Line number within the originating file
        """
        raise NotImplementedError

    @property
    def line_text(self) -> str:
        """
        Source text of the originating line
        """
        raise NotImplementedError

    @property
    def line_selection(self) -> Tuple[int, int]:
        """
        Start/end coordinates of the selection within line_text.
        If the actual selection spans multiple lines, it is clamped to the end
        of this line.
        """
        raise NotImplementedError

#-------------------------------------------------------------------------------
class DirectSourceRef(DetailedFileSourceRef):
    """
    Source reference that points directly to a file's coordinates.
    """
    def __init__(self, path: str, start_idx: int, end_idx: int):
        super().__init__(path)
        self._start_idx = start_idx
        self._end_idx = end_idx

        self._line = None # type: int
        self._line_text = None # type: str
        self._line_selection = None # type: Tuple[int, int]

    def _extract_line_info(self) -> None:
        idx = 0
        lineno = 1
        line_start_idx = 0

        with open(self.path, 'r', newline='', encoding='utf_8') as fp:
            while True:
                line_text = fp.readline()
                assert line_text != ""

                idx += len(line_text)

                # This line contains the start_idx
                if self._start_idx < idx:
                    line_text = line_text.rstrip("\n").rstrip("\r")
                    self._line = lineno
                    self._line_text = line_text
                    start_column = self._start_idx - line_start_idx

                    # Get the end of the selection
                    end_column = self._end_idx - line_start_idx
                    # Clamp it to the end of the current line
                    if end_column >= len(line_text):
                        end_column = len(line_text) - 1

                    self._line_selection = (start_column, end_column)

                    break

                lineno += 1
                line_start_idx = idx

    @property
    def path(self) -> str:
        return self._path

    @property
    def line(self) -> int:
        if self._line is None:
            self._extract_line_info()
        return self._line

    @property
    def line_text(self) -> str:
        if self._line_text is None:
            self._extract_line_info()
        return self._line_text

    @property
    def line_selection(self) -> Tuple[int, int]:
        if self._line_selection is None:
            self._extract_line_info()
        return self._line_selection



class SegmentedSourceRef(DirectSourceRef):
    """
    Source reference that requires coordinate transformation through one or more
    segment maps
    """

    def __init__(self, seg_map: SegmentMap, start_idx: int, end_idx: int):
        super().__init__(None, None, None)

        self._seg_map = seg_map
        self._seg_start_idx = start_idx
        self._seg_end_idx = end_idx

    @property
    def path(self) -> str:
        if self._path is None:
            self._resolve_seg_map()
        return self._path

    def _resolve_seg_map(self) -> None:
        self._start_idx, self._end_idx, self._path = self._seg_map.get_selection(
            self._seg_start_idx, self._seg_end_idx
        )


#-------------------------------------------------------------------------------
def src_ref_from_antlr(antlr_ref: Union[CommonToken, TerminalNodeImpl, ParserRuleContext]) -> 'SourceRefBase':
    # Normalize to pair of CommonToken objects
    if isinstance(antlr_ref, CommonToken):
        token = antlr_ref
        end_token = None
    elif isinstance(antlr_ref, TerminalNodeImpl):
        token = antlr_ref.symbol
        end_token = None
    elif isinstance(antlr_ref, ParserRuleContext):
        # antlr_ref is an entire context (token range)
        token = antlr_ref.start
        end_token = antlr_ref.stop
    else:
        print(antlr_ref)
        raise NotImplementedError

    # Extract selection coordinates
    start = token.start
    if end_token is None:
        end = token.stop
    else:
        end = end_token.stop

    inputStream = token.getInputStream()
    if not isinstance(inputStream, PreprocessedInputStream):
        # This originated from a non-file, so the full src_ref is not known
        # TODO: Eventually extend this to a stream source ref
        return None

    return SegmentedSourceRef(inputStream.seg_map, start, end)



from typing import TYPE_CHECKING, Dict, Optional, Type

from .bases import PropertyRule
from .user_defined import UserProperty, ExternalUserProperty

from .. import rdltypes
from ..core.helpers import get_all_subclasses

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

class PropertyRuleBook:
    def __init__(self, env: 'RDLEnvironment'):
        self.env = env

        # Auto-discover all properties defined below and load into dict
        self.rdl_properties = {} # type: Dict[str, PropertyRule]
        for prop in get_all_subclasses(PropertyRule):
            if prop.__name__.startswith("Prop_"):
                prop_inst = prop(self.env)
                self.rdl_properties[prop_inst.get_name()] = prop(self.env)

        self.user_properties = {} # type: Dict[str, UserProperty]

        self.rdl_prop_refs = {} # type: Dict[str, Type[rdltypes.PropertyReference]]
        for prop_ref in get_all_subclasses(rdltypes.PropertyReference):
            if prop_ref.__name__.startswith("PropRef_"):
                prop_name = prop_ref.get_name()
                self.rdl_prop_refs[prop_name] = prop_ref

    def lookup_property(self, prop_name: str, include_soft_udp: bool=False) -> Optional[PropertyRule]:
        if prop_name in self.rdl_properties:
            return self.rdl_properties[prop_name]
        elif prop_name in self.user_properties:
            udp = self.user_properties[prop_name]
            if isinstance(udp, ExternalUserProperty) and udp.is_soft and not include_soft_udp:
                # Soft UDPs do not officially exist until they are explicitly defined
                return None
            return udp
        else:
            return None

    def lookup_prop_ref_type(self, prop_ref_name):
        # type: (str) -> Optional[Type[rdltypes.PropertyReference]]
        return self.rdl_prop_refs.get(prop_ref_name, None)

    def register_udp(self, udp: UserProperty, src_ref: 'SourceRefBase') -> None:
        if udp.name in self.user_properties:

            existing_udp = self.user_properties[udp.name]
            if isinstance(existing_udp, ExternalUserProperty) and existing_udp.is_soft:
                # Existing UDP is soft. Check if incoming UDP is equivalent
                if existing_udp.bindable_to != udp.bindable_to:
                    self.env.msg.error(
                        "The property definition for the feature extension '%s' uses a different 'components' definition from what this tool expects." % udp.name,
                        src_ref
                    )
                if existing_udp.valid_type != udp.valid_type:
                    self.env.msg.error(
                        "The property definition for the feature extension '%s' uses a different 'type' definition from what this tool expects." % udp.name,
                        src_ref
                    )
                if existing_udp.default_assignment != udp.default_assignment:
                    self.env.msg.error(
                        "The property definition for the feature extension '%s' uses a different 'default' definition from what this tool expects." % udp.name,
                        src_ref
                    )
                if existing_udp.constr_componentwidth != udp.constr_componentwidth:
                    self.env.msg.fatal(
                        "The property definition for the feature extension '%s' uses a different 'constraint' definition from what this tool expects." % udp.name,
                        src_ref
                    )

                # Now that the soft UDP has been explicitly defined by the user,
                # unmark it as soft
                existing_udp.is_soft = False
                return
            self.env.msg.fatal(
                "Multiple declarations of user-defined property '%s'"
                % udp.name,
                src_ref
            )

        if udp.name in self.rdl_properties:
            self.env.msg.fatal(
                "User-defined property '%s' cannot be the same name as a built-in SystemRDL property"
                % udp.name,
                src_ref
            )

        self.user_properties[udp.name] = udp



from typing import TYPE_CHECKING, Any, Optional, Union

from .bases import PropertyRule, PropertyRuleBoolPair

from .. import component as comp
from .. import node as m_node
from ..ast.cast import AssignmentCast
from .. import rdltypes

if TYPE_CHECKING:
    from ..source_ref import SourceRefBase
    from ..compiler import RDLEnvironment

#===============================================================================
# General Properties
#===============================================================================
class Prop_name(PropertyRule):
    """
    Specifies a more descriptive name
    (5.2.1)
    """
    bindable_to = {comp.Addrmap, comp.Field, comp.Mem, comp.Reg, comp.Regfile, comp.Signal}
    valid_types = (str,)
    default = ""
    dyn_assign_allowed = True
    mutex_group = None

    def get_default(self, node: m_node.Node) -> str:
        """
        If name is undefined, it is presumed to be the instance name.
        (5.2.1.1)
        """
        return node.inst_name


class Prop_desc(PropertyRule):
    """
    Describes the component's purpose.
    (5.2.1)
    """
    bindable_to = {comp.Addrmap, comp.Field, comp.Mem, comp.Reg, comp.Regfile, comp.Signal}
    valid_types = (str,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None


class Prop_dontcompare(PropertyRule):
    """
    Indicates the components read data shall be discarded and not compared
    against expected results.
    (5.2.2)
    """
    bindable_to = {comp.Addrmap, comp.Reg, comp.Regfile, comp.Field}
    valid_types = (int, bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "O"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        super().assign_value(comp_def, value, src_ref)

        # int type only makes sense if assigned to a field (since it is a bitmask)
        # If assigned to any other components, exclusively cast it to a boolean
        if not isinstance(comp_def, comp.Field):
            value = comp_def.properties[self.get_name()]
            value = AssignmentCast(self.env, src_ref, value, bool)
            comp_def.properties[self.get_name()] = value

    def validate(self, node: m_node.Node, value: Any) -> None:
        donttest = node.get_property('donttest')

        if isinstance(node, m_node.FieldNode):
            # 5.2.2.1-a: If value is a bit mask, the mask shall have the same width
            # as the field
            if isinstance(value, int):
                if value >= (1 << node.width):
                    self.env.msg.error(
                        "Bit mask (%d) of property 'dontcompare' exceeds width (%d) of field '%s'"
                        % (value, node.width, node.inst_name),
                        self.get_src_ref(node)
                    )

            # Normalize values to masks
            if isinstance(value, bool):
                if value:
                    value = (1 << node.width) - 1
                else:
                    value = 0
            if isinstance(donttest, bool):
                if donttest:
                    donttest = (1 << node.width) - 1
                else:
                    donttest = 0

            # 5.2.2.1-c.2: dontcompare/donttest cannot have one true and the
            # other non-zero
            # 5.2.2.1-c.3: the bitwise AND of dontcompare/donttest  masks
            # shall be zero (0) for a particular component
            # (i.e., donttest & dontcompare = 0)
            if value & donttest:
                self.env.msg.error(
                    "A field's bit cannot have both 'dontcompare' and 'donttest' properties enabled",
                    node.inst.inst_src_ref
                )

        else:
            # A boolean may end up cast as an int. Normalize 0 or 1 to boolean
            if isinstance(value, int) and value in (0, 1):
                value = bool(value)

            # 5.2.2.1-b: can also be applied to reg, regfile, and addrmap
            # components, but only as a boolean
            if not isinstance(value, bool):
                self.env.msg.error(
                    "Property 'dontcompare' expects a boolean for non-field types. Got an integer in '%s'"
                    % (node.inst_name),
                    node.inst.inst_src_ref
                )

            # 5.2.2.1-c.1: dontcompare/donttest cannot both be set to true
            if donttest and value:
                self.env.msg.error(
                    "Properties dontcompare/donttest cannot both be set to true in '%s'"
                    % (node.inst_name),
                    node.inst.inst_src_ref
                )


class Prop_donttest(PropertyRule):
    """
    Indicates the component is not included in structural testing.
    (5.2.2)
    """
    bindable_to = {comp.Addrmap, comp.Reg, comp.Regfile, comp.Field}
    valid_types = (int, bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "O"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        super().assign_value(comp_def, value, src_ref)

        # int type only makes sense if assigned to a field (since it is a bitmask)
        # If assigned to any other components, exclusively cast it to a boolean
        if not isinstance(comp_def, comp.Field):
            value = comp_def.properties[self.get_name()]
            value = AssignmentCast(self.env, src_ref, value, bool)
            comp_def.properties[self.get_name()] = value

    def validate(self, node: m_node.Node, value: Any) -> None:
        if isinstance(node, m_node.FieldNode):
            # 5.2.2.1-a: If value is a bit mask, the mask shall have the same width
            # as the field
            if isinstance(value, int):
                if value >= (1 << node.width):
                    self.env.msg.error(
                        "Bit mask (%d) of property 'donttest' exceeds width (%d) of field '%s'"
                        % (value, node.width, node.inst_name),
                        self.get_src_ref(node)
                    )
        else:
            # A boolean may end up cast as an int. Normalize 0 or 1 to boolean
            if isinstance(value, int) and value in (0, 1):
                value = bool(value)

            # 5.2.2.1-b: can also be applied to reg, regfile, and addrmap
            # components, but only as a boolean
            if not isinstance(value, bool):
                self.env.msg.error(
                    "Property 'donttest' expects a boolean for non-field types. Got an integer in '%s'"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )


class Prop_ispresent(PropertyRule):
    """
    Setting ispresent to false causes the given component instance to be removed
    from the final specification.
    (5.3)
    """
    bindable_to = {comp.Addrmap, comp.Field, comp.Mem, comp.Reg, comp.Regfile, comp.Signal}
    valid_types = (bool,)
    default = True
    dyn_assign_allowed = True
    mutex_group = None


class Prop_errextbus(PropertyRule):
    bindable_to = {comp.Addrmap, comp.Reg, comp.Regfile}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        # 10.6.1-h: errextbus is only valid for external registers
        if (node.inst.external is False) and (value is True):
            self.env.msg.error(
                "The 'errextbus' property is set to 'true', but instance '%s' is not external"
                % (node.inst_name),
                node.inst.inst_src_ref
            )


class Prop_hdl_path(PropertyRule):
    bindable_to = {comp.Addrmap, comp.Reg, comp.Regfile}
    valid_types = (str,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None


class Prop_hdl_path_gate(PropertyRule):
    bindable_to = {comp.Addrmap, comp.Reg, comp.Regfile}
    valid_types = (str,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None


class Prop_hdl_path_gate_slice(PropertyRule):
    bindable_to = {comp.Field, comp.Mem}
    valid_types = (rdltypes.ArrayPlaceholder(str),)
    default = None
    dyn_assign_allowed = True
    mutex_group = None


class Prop_hdl_path_slice(PropertyRule):
    bindable_to = {comp.Field, comp.Mem}
    valid_types = (rdltypes.ArrayPlaceholder(str),)
    default = None
    dyn_assign_allowed = True
    mutex_group = None


#===============================================================================
# Signal Properties
#===============================================================================

class Prop_signalwidth(PropertyRule):
    """
    Width of the signal.
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = False
    mutex_group = None

    def get_default(self, node: m_node.Node) -> Optional[int]:
        """
        If not explicitly set, inherits the instantiation's width
        """
        assert isinstance(node, m_node.SignalNode)
        return node.width


class Prop_sync(PropertyRuleBoolPair):
    """
    Signal is synchronous to the clock of the component.
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = True
    dyn_assign_allowed = True
    mutex_group = "N"

    opposite_property = "async"


class Prop_async(PropertyRuleBoolPair):
    """
    Signal is asynchronous to the clock of the component.
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "N"

    opposite_property = "sync"


class Prop_cpuif_reset(PropertyRule):
    """
    Default signal to use for resetting the software interface logic. If
    cpuif_reset is not defined, this reverts to the default reset signal. This
    parameter only controls the CPU interface of a generated slave.
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        if value is True:
            if not node.get_property('activehigh') and not node.get_property('activelow'):
                self.env.msg.error(
                    "Signal '%s' sets the 'cpuif_reset' property but does not specify whether it is activehigh/activelow"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )



class Prop_field_reset(PropertyRule):
    """
    Default signal to use for resetting field implementations. If field_reset
    is not defined, this reverts to the default reset signal.
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        if value is True:
            if not node.get_property('activehigh') and not node.get_property('activelow'):
                self.env.msg.error(
                    "Signal '%s' sets the 'field_reset' property but does not specify whether it is activehigh/activelow"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )


class Prop_activelow(PropertyRule):
    """
    Signal is active low (state of 0 means ON).
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "A"


class Prop_activehigh(PropertyRule):
    """
    Signal is active high (state of 1 means ON).
    (8.2)
    """
    bindable_to = {comp.Signal}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "A"

#===============================================================================
# Field Properties
#===============================================================================

#-------------------------------------------------------------------------------
# Field access Properties
#-------------------------------------------------------------------------------
class Prop_hw(PropertyRule):
    """
    Design’s ability to sample/update a field.
    (9.4)
    """
    bindable_to = {comp.Field}
    valid_types = (rdltypes.AccessType,)
    default = rdltypes.AccessType.rw
    dyn_assign_allowed = False
    mutex_group = None


class Prop_sw(PropertyRule):
    """
    Programmer’s ability to read/write a field.
    (9.4)
    """
    bindable_to = {comp.Field, comp.Mem}
    valid_types = (rdltypes.AccessType,)
    default = rdltypes.AccessType.rw
    dyn_assign_allowed = True
    mutex_group = None

#-------------------------------------------------------------------------------
# Hardware Signal Properties
#-------------------------------------------------------------------------------
class Prop_next(PropertyRule):
    """
    The next value of the field; the D-input for flip-flops.
    (9.5)
    """
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)

        # 9.5.1-e: next cannot be self-referencing
        if isinstance(value, rdltypes.PropertyReference):
            ref_node = value.node
        else:
            ref_node = value

        if node.get_path() == ref_node.get_path():
            self.env.msg.error(
                "Field '%s' cannot reference itself in next property"
                % (node.inst_name),
                self.get_src_ref(node)
            )

        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

        # Field shall be hardware writable
        # Example in 17.2.8 makes a passing comment that suggests this is a requirement.
        if not node.is_hw_writable:
            self.env.msg.error(
                "Use of the 'next' property requires the field to be hardware-writable.",
                self.get_src_ref(node)
            )


class Prop_reset(PropertyRule):
    """
    The reset value for the field when resetsignal is asserted.
    (9.5)
    """
    bindable_to = {comp.Field}
    valid_types = (int, comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if isinstance(value, int):
            # 9.5.1-c: The reset value cannot be larger than can fit in the field
            if value >= (1 << node.width):
                self.env.msg.error(
                    "The reset value (%d) of field '%s' cannot fit within its width (%d)"
                    % (value, node.inst_name, node.width),
                    self.get_src_ref(node)
                )
        elif isinstance(value, m_node.FieldNode):
            # 9.5.1-e: reset cannot be self-referencing
            if node.get_path() == value.get_path():
                self.env.msg.error(
                    "Field '%s' cannot reference itself in reset property"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )
        elif isinstance(value, (m_node.SignalNode, rdltypes.PropertyReference)):
            pass
        else:
            raise RuntimeError

        # Check width
        # 9.5.1-d: When reset is a reference, it shall reference another
        # field of the same size.
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_resetsignal(PropertyRule):
    """
    Reference to the signal used to reset the field
    (9.5)
    """
    bindable_to = {comp.Field}
    valid_types = (comp.Signal,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(value, m_node.SignalNode)
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)

        if not value.get_property('activehigh') and not value.get_property('activelow'):
            self.env.msg.error(
                "Signal '%s' referenced in 'resetsignal' does not specify whether it is activehigh/activelow"
                % (value.inst_name),
                self.get_src_ref(node)
            )

    def get_default(self, node: m_node.Node) -> Optional[m_node.SignalNode]:
        """
        If no field reset signal was explicitly assigned, search for signals in
        the enclosing hierarchy with field_reset=True
        """
        current_node = node
        while current_node is not None:
            for signal in current_node.signals():
                if signal.get_property('field_reset'):
                    return signal
            current_node = current_node.parent

        return None

#-------------------------------------------------------------------------------
# Software access properties
#-------------------------------------------------------------------------------

class Prop_rclr(PropertyRule):
    """
    Clear on read (field = 0).
    (9.6)6
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "P"

    def get_default(self, node: m_node.Node) -> bool:
        """
        If not explicitly set, check if onread sets the equivalent
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("onread", None) == rdltypes.OnReadType.rclr:
            return True
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True and not node.is_sw_readable:
            self.env.msg.error(
                "Field '%s' sets the 'rclr' property but does not have software read access"
                % (node.inst_name),
                self.get_src_ref(node)
            )


class Prop_rset(PropertyRule):
    """
    Set on read (field = all 1’s).
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "P"

    def get_default(self, node: m_node.Node) -> bool:
        """
        If not explicitly set, check if onread sets the equivalent
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("onread", None) == rdltypes.OnReadType.rset:
            return True
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True and not node.is_sw_readable:
            self.env.msg.error(
                "Field '%s' sets the 'rset' property but does not have software read access"
                % (node.inst_name),
                self.get_src_ref(node)
            )


class Prop_onread(PropertyRule):
    """
    Read side-effect.
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (rdltypes.OnReadType,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "P"

    def get_default(self, node: m_node.Node) -> Optional[rdltypes.OnReadType]:
        """
        If not explicitly set, check if rset or rclr imply the value
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("rset", False):
            return rdltypes.OnReadType.rset
        elif node.inst.properties.get("rclr", False):
            return rdltypes.OnReadType.rclr
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        # 9.6.1-i A field with an onread property shall have software read access
        if (value is not None) and not node.is_sw_readable:
            self.env.msg.error(
                "Field '%s' has an 'onread' property but does not have software read access"
                % (node.inst_name),
                self.get_src_ref(node)
            )

        # 9.6.1-j A field with an onread value of ruser shall be external
        if (node.inst.external is False) and (value == rdltypes.OnReadType.ruser):
            self.env.msg.error(
                "The 'onread' property is set to 'ruser', but instance '%s' is not external"
                % (node.inst_name),
                node.inst.inst_src_ref
            )

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_woclr(PropertyRule):
    """
    Write one to clear (field = field & ~write_data).
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "B"

    def get_default(self, node: m_node.Node) -> bool:
        """
        If not explicitly set, check if onwrite sets the equivalent
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("onwrite", None) == rdltypes.OnWriteType.woclr:
            return True
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True and not node.is_sw_writable:
            self.env.msg.error(
                "Field '%s' sets the 'woclr' property but does not have software write access"
                % (node.inst_name),
                self.get_src_ref(node)
            )


class Prop_woset(PropertyRule):
    """
    Write one to set (field = field | write_data).
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "B"

    def get_default(self, node: m_node.Node) -> bool:
        """
        If not explicitly set, check if onwrite sets the equivalent
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("onwrite", None) == rdltypes.OnWriteType.woset:
            return True
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True and not node.is_sw_writable:
            self.env.msg.error(
                "Field '%s' sets the 'woset' property but does not have software write access"
                % (node.inst_name),
                self.get_src_ref(node)
            )


class Prop_onwrite(PropertyRule):
    """
    Write side-effect
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (rdltypes.OnWriteType,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "B"

    def get_default(self, node: m_node.Node) -> Optional[rdltypes.OnWriteType]:
        """
        If not explicitly set, check if woset or woclr imply the value
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if node.inst.properties.get("woset", False):
            return rdltypes.OnWriteType.woset
        elif node.inst.properties.get("woclr", False):
            return rdltypes.OnWriteType.woclr
        else:
            return self.default

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        # 9.6.1-l A field with an onwrite property shall have software write access.
        if (value is not None) and not node.is_sw_writable:
            self.env.msg.error(
                "Field '%s' has an 'onwrite' property but does not have software write access"
                % (node.inst_name),
                self.get_src_ref(node)
            )

        # 9.6.1-m A field with an onwrite value of wuser shall be external
        if (node.inst.external is False) and (value == rdltypes.OnWriteType.wuser):
            self.env.msg.error(
                "The 'onwrite' property is set to 'wuser', but instance '%s' is not external"
                % (node.inst_name),
                node.inst.inst_src_ref
            )

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def _validate_swwe_writable(env: "RDLEnvironment", node: m_node.Node, prop_name: str, prop_value: Union[bool, rdltypes.ComponentRef]) -> None:
    # swwe and swwel properties enable a hardware signal that allows the
    # writability of a field to change at runtime.
    # If either property is used, then the field is implicitly considered
    # writable at some point.
    # If the field's 'sw' property conflicts with this, emit an error to
    # the user.
    this_f_sw = node.get_property('sw')

    if isinstance(prop_value, rdltypes.ComponentRef):
        shall_be_writable = True
    elif prop_value is True:
        shall_be_writable = True
    else:
        shall_be_writable = False

    if shall_be_writable:
        if this_f_sw == rdltypes.AccessType.r:
            env.msg.error(
                "Field's software access is 'sw=r' but property '%s' implies it can be written in some situations during runtime."
                % prop_name,
                node.inst.property_src_ref.get(prop_name, node.inst.inst_src_ref)
            )
        elif this_f_sw == rdltypes.AccessType.na:
            env.msg.error(
                "Field's software access is 'sw=na' but property '%s' implies it can be written in some situations during runtime."
                % prop_name,
                node.inst.property_src_ref.get(prop_name, node.inst.inst_src_ref)
            )


class Prop_swwe(PropertyRule):
    """
    Override software-writeability of this field.
    Field is writable if signal/field/value is True
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "R"

    def validate(self, node: m_node.Node, value: Any) -> None:
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)
        _validate_swwe_writable(self.env, node, "swwe", value)


class Prop_swwel(PropertyRule):
    """
    Override software-writeability of this field.
    Field is writable if signal/field/value is False
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "R"

    def validate(self, node: m_node.Node, value: Any) -> None:
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)
        _validate_swwe_writable(self.env, node, "swwel", value)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_swmod(PropertyRule):
    """
    Indicates a generated output signal shall notify hardware when this field is
    modified by software
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None


class Prop_swacc(PropertyRule):
    """
    Indicates a generated output signal shall notify hardware when this field is
    accessed by software
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None


class Prop_singlepulse(PropertyRule):
    """
    Field asserts for one cycle when written 1 and then clears back to 0
    on the next cycle
    If set, field shall be instantiated with a width of 1 and the reset value
    shall be specified as 0
    (9.6)
    """
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        # 9.6.1-g: singlepulse fields shall be instantiated with a width of 1
        # and the reset value shall be specified as 0
        if value:
            if node.width != 1:
                self.env.msg.error(
                    "Field '%s' marked as 'singlepulse' shall have width of 1"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )

            if node.get_property('reset') != 0:
                self.env.msg.error(
                    "Field '%s' marked as 'singlepulse' shall have a reset value of 0"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )

            if not node.is_sw_writable:
                self.env.msg.error(
                    "Field '%s' marked as 'singlepulse' shall be writable by software"
                    % (node.inst_name),
                    self.get_src_ref(node)
                )

            # singlepulse does not make sense alongside any onwrite properties
            # that conflict with singlepulse semantics
            onwrite = node.get_property('onwrite')
            if onwrite is not None:
                illegal_onwrite = (
                    rdltypes.OnWriteType.woclr,
                    rdltypes.OnWriteType.wclr,
                )
                if onwrite in illegal_onwrite:
                    self.env.msg.error(
                        "Field '%s' marked as 'singlepulse' has conflicting 'onwrite' value of '%s'"
                        % (node.inst_name, onwrite.name),
                        self.get_src_ref(node)
                    )

#-------------------------------------------------------------------------------
# Hardware access properties
#-------------------------------------------------------------------------------

class Prop_we(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "C"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)

        if isinstance(value, comp.VectorComponent):
            uses_we = True
        else:
            # value is boolean
            uses_we = value

        if uses_we and (node.get_property('hw') not in (rdltypes.AccessType.rw, rdltypes.AccessType.w)):
            self.env.msg.error(
                "Field '%s' sets property 'we', but the field's 'hw' property indicates is not writable by hardware"
                % (node.inst_name),
                self.get_src_ref(node)
            )

        if uses_we and not node.implements_storage:
            self.env.msg.error(
                "Use of 'we' property on field '%s' that does not implement storage does not make sense"
                % (node.inst_name),
                self.get_src_ref(node)
            )


class Prop_wel(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "C"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)

        if isinstance(value, comp.VectorComponent):
            uses_we = True
        else:
            # value is boolean
            uses_we = value

        if uses_we and (node.get_property('hw') not in (rdltypes.AccessType.rw, rdltypes.AccessType.w)):
            self.env.msg.error(
                "Field '%s' sets property 'wel', but the field's 'hw' property indicates is not writable by hardware"
                % (node.inst_name),
                self.get_src_ref(node)
            )

        if uses_we and not node.implements_storage:
            self.env.msg.error(
                "Use of 'wel' property on field '%s' that does not implement storage does not make sense"
                % (node.inst_name),
                self.get_src_ref(node)
            )

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_anded(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None


class Prop_ored(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None


class Prop_xored(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_fieldwidth(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = False
    mutex_group = None

    def get_default(self, node: m_node.Node) -> Optional[int]:
        """
        If not explicitly set, inherits the instantiation's width
        """
        assert isinstance(node, m_node.FieldNode)
        return node.width

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_hwclr(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)


class Prop_hwset(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Prop_hwenable(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "D"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_hwmask(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "D"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


#-------------------------------------------------------------------------------
# Counter field properties
#-------------------------------------------------------------------------------

class Prop_counter(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "E"


class CounterProperty(PropertyRule):
    def validate(self, node: m_node.Node, value: Any) -> None:
        # If using this property, validate that the field was marked as a counter
        if not node.get_property('counter'):
            self.env.msg.error(
                "Field '%s' uses property '%s' which is reserved for counter fields, but the field is not marked as a counter"
                % (node.inst_name, self.get_name()),
                self.get_src_ref(node)
            )


class Prop_threshold(CounterProperty):
    """
    alias of incrthreshold.
    """
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "incrthreshold alias"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        """
        Set both alias and actual value
        """
        super().assign_value(comp_def, value, src_ref)
        comp_def.properties['incrthreshold'] = comp_def.properties['threshold']

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_saturate(CounterProperty):
    """
    alias of incrsaturate.
    """
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "incrsaturate alias"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        """
        Set both alias and actual value
        """
        super().assign_value(comp_def, value, src_ref)
        comp_def.properties['incrsaturate'] = comp_def.properties['saturate']

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_incrthreshold(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "incrthreshold alias"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        """
        Set both alias and actual value
        """
        super().assign_value(comp_def, value, src_ref)
        comp_def.properties['threshold'] = comp_def.properties['incrthreshold']

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_incrsaturate(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "incrsaturate alias"

    def assign_value(self, comp_def: comp.Component, value: Any, src_ref: 'SourceRefBase') -> None:
        """
        Set both alias and actual value
        """
        super().assign_value(comp_def, value, src_ref)
        comp_def.properties['saturate'] = comp_def.properties['incrsaturate']

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_overflow(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)

        if node.get_property('incrsaturate') is not False:
            self.env.msg.error(
                "Use of 'overflow' property is meaningless. Counter sets the "
                "'incrsaturate' property which makes it unable to overflow",
                self.get_src_ref(node)
            )


class Prop_underflow(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)

        if node.get_property('decrsaturate') is not False:
            self.env.msg.error(
                "Use of 'underflow' property is meaningless. Counter sets the "
                "'decrsaturate' property which makes it unable to underflow",
                self.get_src_ref(node)
            )


class Prop_incr(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)


class Prop_incrvalue(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "F"

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_width_eq_or_smaller(node, value)
        self._validate_ref_is_present(node, value)

    def get_default(self, node: m_node.Node) -> Any:
        assert isinstance(node, m_node.FieldNode)
        if node.is_up_counter and node.get_property('incrwidth') is None:
            # Is counter, but no alternatives to increment value were specified.
            return 1

        return None


class Prop_incrwidth(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "F"

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)

        if not 1 <= value <= node.width:
            self.env.msg.error(
                "A counter's 'incrwidth' must be between 1 and the counter's width (%d)"
                % node.width,
                self.get_src_ref(node)
            )


class Prop_decrvalue(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "G"

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_width_eq_or_smaller(node, value)
        self._validate_ref_is_present(node, value)

    def get_default(self, node: m_node.Node) -> Any:
        assert isinstance(node, m_node.FieldNode)
        if node.is_down_counter and node.get_property('decrwidth') is None:
            # Is counter, but no alternatives to decrement value were specified.
            return 1

        return None


class Prop_decr(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width_is_1(node, value)
        self._validate_ref_is_present(node, value)


class Prop_decrwidth(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "G"

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)

        if not 1 <= value <= node.width:
            self.env.msg.error(
                "A counter's 'decrwidth' must be between 1 and the counter's width (%d)"
                % node.width,
                self.get_src_ref(node)
            )


class Prop_decrsaturate(CounterProperty):
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)


class Prop_decrthreshold(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (int, bool, comp.Signal, comp.Field, rdltypes.PropertyReference,)
    default = False
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

#-------------------------------------------------------------------------------
# Field access interrupt properties
#-------------------------------------------------------------------------------

class Prop_intr(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "E"


class Prop_intr_type(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (rdltypes.InterruptType,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    @classmethod
    def get_name_cls(cls) -> str:
        # Interrupt modifier type is a "special" hidden property
        # Intentionally override the property name to something that is impossible
        # to define in RDL and collide with: contains a space!
        return "intr type"

    def get_default(self, node: m_node.Node) -> Any:
        if node.get_property('intr'):
            # If unspecified, interrupt fields are level-sensitive by default
            return rdltypes.InterruptType.level
        return None


class Prop_enable(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "J"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

        if value:
            if not node.get_property('intr'):
                self.env.msg.error(
                    "The 'enable' property can only be used on interrupt fields.",
                    self.get_src_ref(node)
                )


class Prop_mask(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "J"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

        if value:
            if not node.get_property('intr'):
                self.env.msg.error(
                    "The 'mask' property can only be used on interrupt fields.",
                    self.get_src_ref(node)
                )


class Prop_haltenable(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "K"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

        if value:
            if not node.get_property('intr'):
                self.env.msg.error(
                    "The 'haltenable' property can only be used on interrupt fields.",
                    self.get_src_ref(node)
                )


class Prop_haltmask(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (comp.Field, comp.Signal, rdltypes.PropertyReference,)
    default = None
    dyn_assign_allowed = True
    mutex_group = "K"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        self._validate_ref_width(node, value)
        self._validate_ref_is_present(node, value)

        if value:
            if not node.get_property('intr'):
                self.env.msg.error(
                    "The 'haltmask' property can only be used on interrupt fields.",
                    self.get_src_ref(node)
                )


class Prop_sticky(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "I"

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True:
            # 'sticky' property doesnt quite make sense for edge-senstive interrupts
            intr_type = node.get_property('intr type')
            if intr_type in {
                rdltypes.InterruptType.posedge,
                rdltypes.InterruptType.negedge,
                rdltypes.InterruptType.bothedge,
            }:
                self.env.msg.error(
                    "Whole-field stickiness only makes sense in level-senstive interrupts, "
                    "but this field is defined as '%s intr'. "
                    "Did you mean to use the 'stickybit' property instead of 'sticky'?"
                    % intr_type.name,
                    self.get_src_ref(node)
                )

            # Use of we/wel qualifier conflicts with sticky property
            if node.get_property('we'):
                self.env.msg.error(
                    "Use of a hardware write-enable on field '%s' does not make "
                    "sense because it is defined as 'sticky'. Sticky fields already "
                    "implicitly control their hardware write-enable behavior based on the input value."
                    % (node.inst_name),
                    node.inst.property_src_ref.get('we', node.inst.inst_src_ref)
                )
            if node.get_property('wel'):
                self.env.msg.error(
                    "Use of a hardware write-enable on field '%s' does not make "
                    "sense because it is defined as 'sticky'. Sticky fields already "
                    "implicitly control their hardware write-enable behavior based on the input value."
                    % (node.inst_name),
                    node.inst.property_src_ref.get('wel', node.inst.inst_src_ref)
                )

            if not node.is_hw_writable:
                self.env.msg.error(
                    "Sticky fields shall be hardware-writable",
                    self.get_src_ref(node)
                )




class Prop_stickybit(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = True
    mutex_group = "I"

    def get_default(self, node: m_node.Node) -> bool:
        """
        Unless specified otherwise, intr fields are implicitly stickybit
        """
        if node.get_property('intr'):
            # Field is an interrupt
            if node.get_property('sticky'):
                # ... and it was defined as multibit sticky. Not stickybit
                return False
            else:
                # By default, implies stickybit
                return True
        else:
            return False

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        if value is True:
            # Use of we/wel qualifier conflicts with stickybit property
            if node.get_property('we'):
                self.env.msg.error(
                    "Use of a hardware write-enable on field '%s' does not make "
                    "sense because it is defined as 'stickybit'. Stickybit fields already "
                    "implicitly control their hardware write-enable behavior based on the input value."
                    % (node.inst_name),
                    node.inst.property_src_ref.get('we', node.inst.inst_src_ref)
                )
            if node.get_property('wel'):
                self.env.msg.error(
                    "Use of a hardware write-enable on field '%s' does not make "
                    "sense because it is defined as 'stickybit'. Stickybit fields already "
                    "implicitly control their hardware write-enable behavior based on the input value."
                    % (node.inst_name),
                    node.inst.property_src_ref.get('wel', node.inst.inst_src_ref)
                )

            if not node.is_hw_writable:
                self.env.msg.error(
                    "Stickybit fields shall be hardware-writable",
                    self.get_src_ref(node)
                )


#-------------------------------------------------------------------------------
# Misc properties
#-------------------------------------------------------------------------------
class Prop_encode(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (rdltypes.UserEnum,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        assert isinstance(node, m_node.FieldNode)
        # 9.10.1-b: The enumeration’s values shall fit inside the field width.
        enum_max = max(map(int, value))
        if enum_max >= (1 << node.width):
            self.env.msg.error(
                "Field '%s' is not wide enough to encode as enum '%s'"
                % (node.inst_name, value.__name__),
                self.get_src_ref(node)
            )


class Prop_precedence(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (rdltypes.PrecedenceType,)
    default = rdltypes.PrecedenceType.sw
    dyn_assign_allowed = True
    mutex_group = None


class Prop_paritycheck(PropertyRule):
    bindable_to = {comp.Field}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = None

#===============================================================================
# Reg Properties
#===============================================================================

class Prop_regwidth(PropertyRule):
    """
    The bit-width of the register (power of two).
    """
    bindable_to = {comp.Reg}
    valid_types = (int,)
    default = 32
    dyn_assign_allowed = False
    mutex_group = None


class Prop_accesswidth(PropertyRule):
    """
    The minimum software access width (power of two) operation that may be
    performed on the register.
    """
    bindable_to = {comp.Reg}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = True
    mutex_group = None

    def get_default(self, node: m_node.Node) -> int:
        """
        10.6.1.d: The default value of the accesswidth property shall be
        identical to the width of the register.
        """
        return node.get_property('regwidth')

    def validate(self, node: m_node.Node, value: Any) -> None:
        # 10.6.1-c: The value of the accesswidth property shall not exceed the
        # value of the regwidth property
        if value > node.get_property('regwidth'):
            self.env.msg.error(
                "Register '%s' has accesswidth of %d which exceeds its regwidth of %d"
                % (node.inst_name, value, node.get_property('regwidth')),
                self.get_src_ref(node)
            )


class Prop_shared(PropertyRule):
    bindable_to = {comp.Reg}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = None

#===============================================================================
# Mem Properties
#===============================================================================

class Prop_mementries(PropertyRule):
    bindable_to = {comp.Mem}
    valid_types = (int,)
    default = 1
    dyn_assign_allowed = False
    mutex_group = None


class Prop_memwidth(PropertyRule):
    bindable_to = {comp.Mem}
    valid_types = (int,)
    default = 32
    dyn_assign_allowed = False
    mutex_group = None

#===============================================================================
# Register file properties
#===============================================================================

class Prop_alignment(PropertyRule):
    bindable_to = {comp.Addrmap, comp.Regfile}
    valid_types = (int,)
    default = None
    dyn_assign_allowed = False
    mutex_group = None

    # RDL spec claims that if unspecified, the default alignment is based on
    # the registers width.
    # If that is taken at face-value, then it would directly conflict with the
    # 'compact' addressing rules in the situation where accesswidth < regwidth
    # Since the equivalent alignment is already handled by the addressing mode
    # rules, the alignment property's default is intentionally left as None
    # in order to distinguish it as unspecified by the user.


class Prop_sharedextbus(PropertyRule):
    bindable_to = {comp.Addrmap, comp.Regfile}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = None

#===============================================================================
# Address map properties
#===============================================================================

class Prop_bigendian(PropertyRuleBoolPair):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False # Default both to false unless one is explicitly set
    dyn_assign_allowed = True
    mutex_group = "L"

    opposite_property = "littleendian"


class Prop_littleendian(PropertyRuleBoolPair):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False # Default both to false unless one is explicitly set
    dyn_assign_allowed = True
    mutex_group = "L"

    opposite_property = "bigendian"


class Prop_addressing(PropertyRule):
    bindable_to = {comp.Addrmap}
    valid_types = (rdltypes.AddressingType,)
    default = rdltypes.AddressingType.regalign
    dyn_assign_allowed = False
    mutex_group = None


class Prop_rsvdset(PropertyRule):
    """
    If true, the read value of all fields not explicitly defined is set to 1
    otherwise, it is set to 0.
    """
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = "Q"


class Prop_rsvdsetX(PropertyRule):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = "Q"


class Prop_msb0(PropertyRuleBoolPair):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = "M"

    opposite_property = "lsb0"


class Prop_lsb0(PropertyRuleBoolPair):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = True
    dyn_assign_allowed = False
    mutex_group = "M"

    opposite_property = "msb0"

#-------------------------------------------------------------------------------
class Prop_bridge(PropertyRule):
    bindable_to = {comp.Addrmap}
    valid_types = (bool,)
    default = False
    dyn_assign_allowed = False
    mutex_group = None

    def validate(self, node: m_node.Node, value: Any) -> None:
        # 13.5: Bridge can only be applied to the root address map
        if value:
            # is bridge
            if (node.parent is not None) and not isinstance(node.parent, m_node.RootNode):
                self.env.msg.error(
                    "The 'bridge' property can only be applied to the root address map.",
                    self.get_src_ref(node)
                )



from typing import TYPE_CHECKING, Any

from .helpers import get_ID_text

from ..parser.SystemRDLParser import SystemRDLParser
from ..parser.SystemRDLVisitor import SystemRDLVisitor
from ..source_ref import src_ref_from_antlr

from .. import rdltypes
from .. import component as comp

if TYPE_CHECKING:
    from ..compiler import RDLCompiler
    from antlr4.Token import CommonToken
    from antlr4 import ParserRuleContext
    from typing import Union, Type

class BaseVisitor(SystemRDLVisitor):

    def __init__(self, compiler: 'RDLCompiler') -> None:
        self.compiler = compiler
        self.msg = compiler.env.msg

    #---------------------------------------------------------------------------
    # Type Handling
    #---------------------------------------------------------------------------
    _DataType_Map = {
        SystemRDLParser.BIT_kw              : int,
        SystemRDLParser.NUMBER_kw           : int,
        SystemRDLParser.LONGINT_kw          : int,
        SystemRDLParser.ACCESSTYPE_kw       : rdltypes.AccessType,
        SystemRDLParser.ADDRESSINGTYPE_kw   : rdltypes.AddressingType,
        SystemRDLParser.ONREADTYPE_kw       : rdltypes.OnReadType,
        SystemRDLParser.ONWRITETYPE_kw      : rdltypes.OnWriteType,
        SystemRDLParser.STRING_kw           : str,
        SystemRDLParser.BOOLEAN_kw          : bool,
        SystemRDLParser.REF_kw              : rdltypes.references.RefType,
        SystemRDLParser.ADDRMAP_kw          : comp.Addrmap,
        SystemRDLParser.REGFILE_kw          : comp.Regfile,
        SystemRDLParser.MEM_kw              : comp.Mem,
        SystemRDLParser.REG_kw              : comp.Reg,
        SystemRDLParser.FIELD_kw            : comp.Field,
        SystemRDLParser.SIGNAL_kw           : comp.Signal,
    }
    def datatype_from_token(self, token):
        # type: ('CommonToken') -> Type[Union[int, str, bool, rdltypes.BuiltinEnum, rdltypes.UserEnum, rdltypes.UserStruct, rdltypes.references.RefType, comp.Component]]
        """
        Given a SystemRDLParser token, lookup the type
        This only includes types under the "data_type" grammar rule
        """

        if token.type == SystemRDLParser.ID:
            # Is an identifier for either an enum or struct type

            typ = self.compiler.namespace.lookup_type(get_ID_text(token))
            if typ is None:
                self.msg.fatal(
                    "Type '%s' is not defined" % get_ID_text(token),
                    src_ref_from_antlr(token)
                )

            if rdltypes.is_user_enum(typ) or rdltypes.is_user_struct(typ):
                return typ # type: ignore
            else:
                self.msg.fatal(
                    "Type '%s' is not a struct or enum" % get_ID_text(token),
                    src_ref_from_antlr(token)
                )

        return self._DataType_Map[token.type]

    #---------------------------------------------------------------------------
    # Keyword passthrough visitors
    #---------------------------------------------------------------------------

    # It is convenient to be able to group commonly-used sets of tokens in the
    # grammar.
    # These visitors propagate the original tokens all the way back up to the
    # visitor that actually needs to know which keyword was used.

    def passthru_kw_token(self, ctx: 'ParserRuleContext') -> Any:
        if ctx.kw is not None:
            return ctx.kw
        else:
            return self.visitChildren(ctx)

    def visitComponent_inst_type(self, ctx: SystemRDLParser.Component_inst_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitComponent_type(self, ctx: SystemRDLParser.Component_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitComponent_type_primary(self, ctx: SystemRDLParser.Component_type_primaryContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitData_type(self, ctx: SystemRDLParser.Data_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitBasic_data_type(self, ctx: SystemRDLParser.Basic_data_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitProp_keyword(self, ctx: SystemRDLParser.Prop_keywordContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitProp_mod(self, ctx: SystemRDLParser.Prop_modContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitUdp_data_type(self, ctx: SystemRDLParser.Udp_data_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitUdp_comp_type(self, ctx: SystemRDLParser.Udp_comp_typeContext) -> Any:
        return self.passthru_kw_token(ctx)

    def visitEval_expr_root(self, ctx:SystemRDLParser.Eval_expr_rootContext) -> Any:
        return self.visit(ctx.expr()) # type: ignore



import re
from typing import TYPE_CHECKING, Optional

from . import helpers
from ..node import Node, AddressableNode

if TYPE_CHECKING:
    from markdown import Markdown

def rdlfc_to_html(text: str, node: Optional[Node]=None, md: Optional['Markdown']=None, is_desc: bool=True) -> str:
    """
    Convert an RDLFormatCode string to HTML
    """

    # --------------------------------------------------------------------------
    # Remove any common indentation
    # --------------------------------------------------------------------------
    text = helpers.dedent_text(text)

    # --------------------------------------------------------------------------
    # Parse and replace RDLFormatCode Tags
    # --------------------------------------------------------------------------
    token_spec = [
        ('b', r'\[b\]'),
        ('xb', r'\[/b\]'),
        ('i', r'\[i\]'),
        ('xi', r'\[/i\]'),
        ('u', r'\[u\]'),
        ('xu', r'\[/u\]'),
        ('color', r'\[color=[^\]]+\]'),
        ('xcolor', r'\[/color\]'),
        ('size', r'\[size=[^\]]+\]'),
        ('xsize', r'\[/size\]'),
        ('plainurl', r'\[url\].*?\[/url\]'),
        ('url', r'\[url=[^\]]+\]'),
        ('xurl', r'\[/url\]'),
        ('email', r'\[email\].*?\[/email\]'),
        ('img', r'\[img\].*?\[/img\]'),
        ('code', r'\[code\].*?\[/code\]'),
        ('list', r'\[list(?:=[^\]]+)?\]'),
        ('bullet', r'\[\*\]'),
        ('xlist', r'\[/list\]'),
        ('quote', r'\[quote\]'),
        ('xquote', r'\[/quote\]'),
        ('br', r'\[br\]'),
        ('lb', r'\[lb\]'),
        ('rb', r'\[rb\]'),
        ('p', r'\[p\]'),
        ('xp', r'\[/p\]'),
        ('sp', r'\[sp\]'),
        ('index', r'\[index\]'),
        ('index_parent', r'\[index_parent\]'),
        ('name', r'\[name\]'),
        ('instname', r'\[instname\]'),
    ]

    if is_desc:
        tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_spec)
    else:
        # filter out tags that are not to be interpreted in 'name' properties
        skipset = {'img', 'list', 'bullet', 'xlist', 'br', 'p', 'xp', 'index', 'index_parent', 'name'}
        tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_spec if pair[0] not in skipset)

    pos = 0
    text_segs = []
    is_first_bullet = []
    list_end_tag = []
    for m in re.finditer(tok_regex, text, re.DOTALL):
        start = m.start()
        end = m.end()

        # Emit prior text
        if start != pos:
            text_segs.append(text[pos:start])
        pos = end

        if m.lastgroup == 'b':
            text_segs.append("<b>")
        elif m.lastgroup == 'xb':
            text_segs.append("</b>")
        elif m.lastgroup == 'i':
            text_segs.append("<i>")
        elif m.lastgroup == 'xi':
            text_segs.append("</i>")
        elif m.lastgroup == 'u':
            text_segs.append("<u>")
        elif m.lastgroup == 'xu':
            text_segs.append("</u>")
        elif m.lastgroup == 'color':
            m2 = re.match(r'\[color=([^\]]+)\]', m.group(0))
            text_segs.append('<span style="color:%s">' % m2.group(1))
        elif m.lastgroup == 'xcolor':
            text_segs.append('</span>')
        elif m.lastgroup == 'size':
            m2 = re.match(r'\[size=([^\]]+)\]', m.group(0))
            text_segs.append('<span style="font-size:%s">' % m2.group(1))
        elif m.lastgroup == 'xsize':
            text_segs.append('</span>')
        elif m.lastgroup == 'plainurl':
            m2 = re.match(r'\[url\](.*?)\[/url\]', m.group(0), re.DOTALL)
            text_segs.append('<a href="%s">%s</a>' % (m2.group(1).strip(), m2.group(1).strip()))
        elif m.lastgroup == 'url':
            m2 = re.match(r'\[url=([^\]]+)\]', m.group(0))
            text_segs.append('<a href="%s">' % m2.group(1).strip())
        elif m.lastgroup == 'xurl':
            text_segs.append('</a>')
        elif m.lastgroup == 'email':
            m2 = re.match(r'\[email\](.*?)\[/email\]', m.group(0), re.DOTALL)
            text_segs.append('<a href="mailto:%s">%s</a>' % (m2.group(1).strip(), m2.group(1).strip()))
        elif m.lastgroup == 'img':
            m2 = re.match(r'\[img\](.*?)\[/img\]', m.group(0), re.DOTALL)
            text_segs.append('<img src="%s">' % m2.group(1))
        elif m.lastgroup == 'code':
            m2 = re.match(r'\[code\](.*?)\s*\[/code\]', m.group(0), re.DOTALL)
            text_segs.append('<code>%s</code>' % m2.group(1))

        elif m.lastgroup == 'list':
            # List start tag
            m2 = re.match(r'\[list(?:=([^\]]+))?\]', m.group(0))
            ltype = m2.group(1)
            if ltype is None:
                text_segs.append('<ul>')
                is_first_bullet.append(True)
                list_end_tag.append('</ul>')
            elif ltype.strip() in ("1", "A", "a", "I", "i"):
                text_segs.append('<ol type="%s">' % ltype.strip())
                is_first_bullet.append(True)
                list_end_tag.append('</ol>')
            else:
                # Bad type. re-emit erroneous list tag
                text_segs.append(m.group(0))

        elif m.lastgroup == 'bullet':
            if len(is_first_bullet) == 0: #pylint: disable=len-as-condition
                # Not inside a list tag. Re-emit erroneous tag
                text_segs.append("\\[\\*\\]")
            else:
                if not is_first_bullet[-1]:
                    text_segs.append("</li>")
                is_first_bullet[-1] = False
                text_segs.append("<li>")

        elif m.lastgroup == 'xlist':
            if len(list_end_tag) == 0: #pylint: disable=len-as-condition
                # Not inside a list tag. Re-emit erroneous tag
                text_segs.append(m.group(0))
            else:
                if not is_first_bullet[-1]:
                    text_segs.append("</li>")
                text_segs.append(list_end_tag[-1])
                is_first_bullet.pop()
                list_end_tag.pop()

        elif m.lastgroup == 'quote':
            text_segs.append('"')
        elif m.lastgroup == 'xquote':
            text_segs.append('"')
        elif m.lastgroup == 'br':
            text_segs.append("<br>")
        elif m.lastgroup == 'lb':
            text_segs.append("[")
        elif m.lastgroup == 'rb':
            text_segs.append("]")
        elif m.lastgroup == 'p':
            text_segs.append("\n\n<p>")
        elif m.lastgroup == 'xp':
            text_segs.append("</p>")
        elif m.lastgroup == 'sp':
            text_segs.append("&nbsp;")
        elif m.lastgroup == 'index':
            if (node is not None) and isinstance(node, AddressableNode) and node.is_array:
                subscripts = []
                if node.current_idx is None:
                    # Index is not known. Use range
                    for dim in node.array_dimensions:
                        subscripts.append("[0:%d]" % (dim-1))
                else:
                    # Index is known
                    for idx in node.current_idx:
                        subscripts.append("[%d]" % idx)
                range_str = "".join(subscripts)
                text_segs.append("<span class='rdlfc-index'>%s</span>" % range_str)
            else:
                text_segs.append(m.group(0))
        elif m.lastgroup == 'index_parent':
            if (node is not None) and (node.parent is not None) and isinstance(node.parent, AddressableNode) and node.parent.is_array:
                subscripts = []
                if node.parent.current_idx is None:
                    # Index is not known. Use range
                    for dim in node.parent.array_dimensions:
                        subscripts.append("[0:%d]" % (dim-1))
                else:
                    # Index is known
                    for idx in node.parent.current_idx:
                        subscripts.append("[%d]" % idx)
                range_str = "".join(subscripts)
                text_segs.append("<span class='rdlfc-index_parent'>%s</span>" % range_str)
            else:
                text_segs.append(m.group(0))
        elif m.lastgroup == 'name':
            if node is not None:
                text_segs.append(node.get_property('name'))
            else:
                text_segs.append(m.group(0))
        elif m.lastgroup == 'instname':
            if node is not None:
                text_segs.append(node.inst_name)
            else:
                text_segs.append(m.group(0))

    # Emit trailing text
    text_segs.append(text[pos:])

    text_out = "".join(text_segs)

    #---------------------------------------------------------------------------
    # Pass through markdown processor
    #---------------------------------------------------------------------------
    if is_desc:
        if md is None:
            # Lazy import markdown
            import markdown # pylint: disable=import-outside-toplevel
            md = markdown.Markdown()
        text_out = md.reset().convert(text_out)

    return text_out



import sys
import os
import platform
import fnmatch
import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

# Replace relative image path with github-hosted one
long_description = long_description.replace(
    "docs/img/overview.svg",
    "https://raw.githubusercontent.com/SystemRDL/systemrdl-compiler/main/docs/img/overview.svg?sanitize=true"
)

with open(os.path.join("systemrdl", "__about__.py"), encoding='utf-8') as f:
    v_dict = {}
    exec(f.read(), v_dict)
    rdl_version = v_dict['__version__']

target = platform.system().lower()
PLATFORMS = {'windows', 'linux', 'darwin', 'cygwin'}
for known in PLATFORMS:
    if target.startswith(known):
        target = known


def run_setup(with_binary):
    if with_binary:

        extra_compile_args = {
            'windows': ['/DANTLR4CPP_STATIC', '/Zc:__cplusplus', '/std:c++17'],
            'linux': ['-std=c++17'],
            'darwin': ['-std=c++17'],
            'cygwin': ['-std=c++17'],
        }

        # Define an Extension object that describes the Antlr accelerator
        parser_ext = setuptools.Extension(
            name='systemrdl.parser.sa_systemrdl_cpp_parser',

            # Add the Antlr runtime source directory to the include search path
            include_dirs=["systemrdl/parser/ext/antlr4-cpp-runtime"],

            # Rather than listing each C++ file (Antlr has a lot!), discover them automatically
            sources=get_files("systemrdl/parser/ext", "*.cpp"),
            depends=get_files("systemrdl/parser/ext", "*.h"),

            extra_compile_args=extra_compile_args.get(target, [])
        )
        ext_modules = [parser_ext]
    else:
        ext_modules = []

    setuptools.setup(
        name="systemrdl-compiler",
        version=rdl_version,
        author="Alex Mykyta",
        author_email="amykyta3@github.com",
        description="Parse and elaborate front-end for SystemRDL 2.0",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/SystemRDL/systemrdl-compiler",
        packages=setuptools.find_packages(exclude=["test"]),
        include_package_data=True,
        ext_modules=ext_modules,
        cmdclass={"build_ext": ve_build_ext},
        python_requires='>=3.5.2',
        install_requires=[
            "antlr4-python3-runtime >= 4.11, < 4.12",
            "colorama",
            "markdown",
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3 :: Only",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
            "Topic :: Software Development :: Compilers",
            "Topic :: Software Development :: Code Generators",
        ],
        project_urls={
            "Documentation": "http://systemrdl-compiler.readthedocs.io",
            "Source": "https://github.com/SystemRDL/systemrdl-compiler",
            "Tracker": "https://github.com/SystemRDL/systemrdl-compiler/issues",
        },
    )


#===============================================================================
from setuptools.command.build_ext import build_ext

def get_files(path, pattern):
    """
    Recursive file search that is compatible with python3.4 and older
    """
    matches = []
    for root, _, filenames in os.walk(path):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


class BuildFailed(Exception):
    pass


class ve_build_ext(build_ext):
    """
    This class extends setuptools to fail with a common BuildFailed exception
    if a build fails
    """

    def run(self):
        try:
            build_ext.run(self)
        except Exception:
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except Exception:
            raise BuildFailed()


# Detect if an alternate interpreter is being used
is_jython = "java" in sys.platform
is_pypy = hasattr(sys, "pypy_version_info")

# Antlr accelerator is no longer supported on older Python versions
is_old_python = sys.version_info[0:2] <= (3, 5)

# Force using fallback python parser under some conditions
using_fallback = is_jython or is_pypy or is_old_python

if not using_fallback:
    try:
        run_setup(with_binary=True)
    except BuildFailed:
        if 'SYSTEMRDL_REQUIRE_BINARY_BUILD' in os.environ:
            # Force failure if binary build is required
            raise
        else:
            using_fallback = True

if using_fallback:
    run_setup(with_binary=False)



from typing import List, TYPE_CHECKING, Tuple, Any, Optional

from ..parser.SystemRDLParser import SystemRDLParser

from .BaseVisitor import BaseVisitor
from .ExprVisitor import ExprVisitor
from .helpers import get_ID_text
from . import helpers

from ..ast import AssignmentCast
from ..source_ref import src_ref_from_antlr
from .. import rdltypes

if TYPE_CHECKING:
    from antlr4.Token import CommonToken
    from ..source_ref import SourceRefBase
    from ..compiler import RDLCompiler
    from .. import component as comp
    from typing import Type

class EnumVisitor(BaseVisitor):

    def __init__(self, compiler: 'RDLCompiler', parent_component: 'comp.Component') -> None:
        super().__init__(compiler)
        self.parent_component = parent_component

    def visitEnum_def(self, ctx):
        # type: (SystemRDLParser.Enum_defContext) -> Tuple[Type[rdltypes.UserEnum], str, SourceRefBase]
        self.compiler.namespace.enter_scope()

        # Allowing enum definitions to access Parameters from the parent scope
        # opens a whole can of worms since there is no clear way for me to ensure
        # they are elaborated accurately (deepcopying a python Enum is a pain)
        # Since this is an extreme corner case, simply disallow it.
        self.compiler.namespace.parent_parameters_visible = False

        enum_name = get_ID_text(ctx.ID())

        # Collect members
        entry_values = [] # type: List[int]
        entry_names = [] # type: List[str]
        members = [] # type: List[rdltypes.UserEnumMemberContainer]
        for enum_entry_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Enum_entryContext):
            name_token, value_expr_ctx, rdl_name, rdl_desc = self.visit(enum_entry_ctx)

            entry_name = get_ID_text(name_token)
            if entry_name in entry_names:
                self.msg.fatal(
                    "Entry '%s' has already been defined in this enum" % entry_name,
                    src_ref_from_antlr(name_token)
                )
            entry_names.append(entry_name)

            if value_expr_ctx is not None:
                # explicit enumerator assignment

                visitor = ExprVisitor(self.compiler)
                expr = visitor.visit(value_expr_ctx)
                expr = AssignmentCast(self.compiler.env, value_expr_ctx, expr, int)
                expr.predict_type()

                # OK to immediately evaluate the expression since there is no way that it
                # can depend on any external references
                entry_value = expr.get_value()
            else:
                # automatic enumerator assignment
                if not entry_values:
                    entry_value = 0
                else:
                    entry_value = entry_values[-1] + 1

            if entry_value in entry_values:
                # Value was already assigned
                self.msg.fatal(
                    "Enumeration encoding values must be unique",
                    src_ref_from_antlr(name_token)
                )

            entry_values.append(entry_value)

            members.append(
                rdltypes.UserEnumMemberContainer(
                    entry_name, entry_value, rdl_name, rdl_desc
                )
            )


        # Create Enum type
        enum_type = rdltypes.UserEnum.define_new(enum_name, members, self.parent_component)

        self.compiler.namespace.exit_scope()
        self.compiler.namespace.parent_parameters_visible = True # restore parameter behavior
        return enum_type, get_ID_text(ctx.ID()), src_ref_from_antlr(ctx.ID())

    def visitEnum_entry(self, ctx: SystemRDLParser.Enum_entryContext) -> Tuple['CommonToken', SystemRDLParser.ExprContext, Optional[str], Optional[str]]:
        name_token = ctx.ID()
        value_expr_ctx = ctx.expr()

        rdl_name = None # type: Optional[str]
        rdl_desc = None # type: Optional[str]

        for pa_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Enum_prop_assignContext):
            prop_token, prop_value = self.visit(pa_ctx)
            prop_name = get_ID_text(prop_token)

            if prop_name == "desc":
                if rdl_desc is not None:
                    self.msg.error(
                        "Property 'desc' was already assigned in this scope",
                        src_ref_from_antlr(prop_token)
                    )
                    continue
                if self.compiler.env.dedent_desc:
                    rdl_desc = helpers.dedent_text(prop_value)
                else:
                    rdl_desc = prop_value
            elif prop_name == "name":
                if rdl_name is not None:
                    self.msg.error(
                        "Property 'name' was already assigned in this scope",
                        src_ref_from_antlr(prop_token)
                    )
                    continue
                rdl_name = prop_value
            else:
                self.msg.fatal(
                    "Illegal enum property assignment '%s'" % prop_name,
                    src_ref_from_antlr(prop_token)
                )

        return name_token, value_expr_ctx, rdl_name, rdl_desc

    def visitEnum_prop_assign(self, ctx: SystemRDLParser.Enum_prop_assignContext) -> Tuple['CommonToken', Any]:
        prop_token = ctx.ID()

        visitor = ExprVisitor(self.compiler)
        prop_expr = visitor.visit(ctx.expr())
        prop_expr = AssignmentCast(self.compiler.env, ctx.expr(), prop_expr, str)
        prop_expr.predict_type()

        # OK to immediately evaluate the expression since there is no way that it
        # can depend on any external references
        prop_value = prop_expr.get_value()

        return prop_token, prop_value



from .__about__ import __version__

from .compiler import RDLCompiler
from .importer import RDLImporter
from .walker import RDLListener, RDLWalker, WalkerAction
from .messages import RDLCompileError

from .node import AddressableNode, VectorNode, SignalNode
from .node import FieldNode, RegNode, RegfileNode, AddrmapNode, MemNode

from .component import AddressableComponent, VectorComponent, Signal
from .component import Field, Reg, Regfile, Addrmap, Mem



from collections import OrderedDict
from typing import TYPE_CHECKING, Optional, Type, Dict, Tuple, List, Any

from .ast_node import ASTNode
from .conditional import is_castable

from .. import rdltypes
from .. import component as comp

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from rdltypes.typing import PreElabRDLType, RDLValue

    OptionalSourceRef = Optional[SourceRefBase]

class BoolLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', val: bool):
        super().__init__(env, src_ref)
        self.val = val

    def predict_type(self) -> Type[bool]:
        return bool

    def get_min_eval_width(self) -> int:
        return 1

    def get_value(self, eval_width: Optional[int]=None) -> bool:
        return self.val


class IntLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', val: int, width: int=64):
        super().__init__(env, src_ref)
        self.val = val
        self.width = width

    def predict_type(self) -> Type[int]:
        return int

    def get_min_eval_width(self) -> int:
        return self.width

    def get_value(self, eval_width: Optional[int]=None) -> int:
        return self.val


class BuiltinEnumLiteral(ASTNode):
    """
    ASTNode wrapper for builtin RDL enumeration types:
    AccessType, OnReadType, OnWriteType, AddressingType, PrecedenceType
    """
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', val: rdltypes.BuiltinEnum):
        super().__init__(env, src_ref)
        self.val = val

    def predict_type(self) -> Type[rdltypes.BuiltinEnum]:
        return type(self.val)

    def get_value(self, eval_width: Optional[int]=None) -> rdltypes.BuiltinEnum:
        return self.val


class EnumLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', val: rdltypes.UserEnum):
        super().__init__(env, src_ref)
        self.val = val

    def predict_type(self) -> Type[rdltypes.UserEnum]:
        return type(self.val)

    def get_min_eval_width(self) -> int:
        return 64

    def get_value(self, eval_width: Optional[int]=None) -> rdltypes.UserEnum:
        return self.val


class StructLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', struct_type: Type[rdltypes.UserStruct], values: Dict[str, Tuple[ASTNode, 'OptionalSourceRef']]):
        super().__init__(env, src_ref)
        self.struct_type = struct_type
        # values is a dict of member_name : (member_expr, member_name_src_ref)
        self.values = values

    def predict_type(self) -> Type[rdltypes.UserStruct]:
        for member_name, (member_expr, member_name_src_ref) in self.values.items():
            member_type = member_expr.predict_type()
            if not is_castable(member_type, self.struct_type._members[member_name]):
                self.msg.fatal(
                    "Expression for member '%s' is not compatible with the expected type"
                    % member_name,
                    member_expr.src_ref
                )

        return self.struct_type

    def get_value(self, eval_width: Optional[int]=None) -> rdltypes.UserStruct:
        resolved_values = OrderedDict()
        for member_name, (member_expr, member_name_src_ref) in self.values.items():
            resolved_values[member_name] = member_expr.get_value()

        return self.struct_type(resolved_values)


class StringLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', val: str):
        super().__init__(env, src_ref)
        self.val = val

    def predict_type(self) -> Type[str]:
        return str

    def get_value(self, eval_width: Optional[int]=None) -> str:
        return self.val


class ArrayLiteral(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', elements: List[ASTNode]):
        super().__init__(env, src_ref)
        self.elements = elements

    def predict_type(self) -> rdltypes.ArrayPlaceholder:

        if not self.elements:
            # Empty array. Element type is indeterminate
            return rdltypes.ArrayPlaceholder(None)

        # Get type of first element
        element_iter = iter(self.elements)
        uniform_type = next(element_iter).predict_type()

        # RDL does not allow directly nested arrays
        assert not isinstance(uniform_type, rdltypes.ArrayPlaceholder)

        # All elements of the array shall have a uniform type
        for element in element_iter:
            this_type = element.predict_type()

            # RDL does not allow directly nested arrays
            assert not isinstance(this_type, rdltypes.ArrayPlaceholder)

            # First check if it is a direct match
            if uniform_type == this_type:
                continue

            # ... nope. Check if it is a compatible ref type
            if uniform_type == rdltypes.references.RefType:
                # array accepts any reference
                if issubclass(this_type, (comp.Component, rdltypes.PropertyReference)):
                    continue

            # ... nope. Check if the array's element type can be generalized as a ref type
            if (
                issubclass(uniform_type, (comp.Component, rdltypes.PropertyReference))
                and issubclass(this_type, (comp.Component, rdltypes.PropertyReference))
            ):
                # Both are still references, so convert this array to a general ref array
                uniform_type = rdltypes.references.RefType
                continue

            # In the event that this is an array of user structs, check if maybe
            # one is a subclass of the other
            if rdltypes.is_user_struct(uniform_type) and rdltypes.is_user_struct(this_type):
                if issubclass(this_type, uniform_type):
                    # element is a specialization of the current uniform type.
                    # all good!
                    continue
                if issubclass(uniform_type, this_type):
                    # element is a more primitive base struct. Make the array more general
                    uniform_type = this_type
                    continue

            self.msg.fatal(
                "Elements of an array shall be the same type",
                self.src_ref
            )

        return rdltypes.ArrayPlaceholder(uniform_type)

    def get_value(self, eval_width: Optional[int]=None) -> List[Any]:
        result = []
        for element in self.elements:
            result.append(element.get_value())
        return result


class ExternalLiteral(ASTNode):
    """
    ASTNode wrapper for literal value that was not compiled from a source file.
    The value provided is not an expression, but the actual value.
    """
    def __init__(self, env: 'RDLEnvironment', value: 'RDLValue'):
        super().__init__(env, None)
        self.value = value

    def predict_type(self) -> 'PreElabRDLType':
        return rdltypes.get_rdltype(self.value)

    def get_min_eval_width(self) -> int:
        if isinstance(self.value, bool):
            return 1
        elif isinstance(self.value, int):
            return 64
        elif rdltypes.is_user_enum(self.value):
            return 64
        else:
            raise RuntimeError

    def get_value(self, eval_width: Optional[int]=None) -> 'RDLValue':
        return self.value



from typing import TYPE_CHECKING, Optional, Type, Any

from .ast_node import ASTNode

from ..core.helpers import truncate_int
from .. import rdltypes
from .. import component as comp

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from rdltypes.typing import PreElabRDLType

    OptionalSourceRef = Optional[SourceRefBase]

#-------------------------------------------------------------------------------
# Width cast operator
# the cast type informs the parser what width to cast to
# The cast width determines the result's width
# Also influences the min eval width of the value expression
class WidthCast(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', v: ASTNode, w_expr: Optional[ASTNode]=None, w_int: int=64):
        super().__init__(env, src_ref)

        if w_expr is not None:
            self.v = v
            self.w_expr = w_expr
            self.cast_width = None
        else:
            self.v = v
            self.w_expr = None
            self.cast_width = w_int

    def predict_type(self) -> Type[int]:
        if self.cast_width is None:
            if not is_castable(self.w_expr.predict_type(), int):
                self.msg.fatal(
                    "Width operand of cast expression is not a compatible numeric type",
                    self.w_expr.src_ref
                )
        if not is_castable(self.v.predict_type(), int):
            self.msg.fatal(
                "Value operand of cast expression cannot be cast to an integer",
                self.v.src_ref
            )

        return int

    def get_min_eval_width(self) -> int:
        if self.cast_width is None:
            self.cast_width = int(self.w_expr.get_value())
        return self.cast_width


    def get_value(self, eval_width: Optional[int]=None) -> int:
        # Truncate to cast width instead of eval width
        if self.cast_width is None:
            self.cast_width = int(self.w_expr.get_value())
        if self.cast_width == 0:
            self.msg.fatal(
                "Cannot cast to width of zero",
                self.src_ref
            )

        eval_width = max(self.cast_width, self.v.get_min_eval_width())
        n = int(self.v.get_value(eval_width))

        return truncate_int(n, self.cast_width)

#-------------------------------------------------------------------------------
# Boolean cast operator

class BoolCast(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', n: ASTNode):
        super().__init__(env, src_ref)
        self.n = n

    def predict_type(self) -> Type[bool]:
        if not is_castable(self.n.predict_type(), bool):
            self.msg.fatal(
                "Value operand of cast expression cannot be cast to a boolean",
                self.src_ref
            )
        return bool

    def get_min_eval_width(self) -> int:
        return 1

    def get_value(self, eval_width: Optional[int]=None) -> bool:
        n = int(self.n.get_value())
        return n != 0

#-------------------------------------------------------------------------------
# Assignment cast
# This is a wrapper expression that normalizes the expression result
# to the expected data type
# This wrapper forces the operand to be evaluated in a self-determined context
# The cast type has no effect on expression evaluation
# During post-compile:
#   Checks that the expression result is of a compatible type
#
# When getting value:
#   Ensures that the expression result gets converted to the resulting type
class AssignmentCast(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', v: ASTNode, dest_type: 'PreElabRDLType'):
        super().__init__(env, src_ref)

        self.v = v
        self.dest_type = dest_type

    def predict_type(self) -> 'PreElabRDLType':
        op_type = self.v.predict_type()

        if not is_castable(op_type, self.dest_type):
            self.msg.fatal(
                "Result of expression is not compatible with the expected type",
                self.src_ref
            )

        return self.dest_type

    def get_min_eval_width(self) -> int:
        return self.v.get_min_eval_width()

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        v = self.v.get_value()

        if self.dest_type == bool:
            return bool(v)
        elif self.dest_type == int:
            return int(v)
        else:
            return v


#===============================================================================

def is_castable(src: Any, dst: Any) -> bool:
    """
    Check if src type can be cast to dst type
    """
    if ((src in [int, bool]) or rdltypes.is_user_enum(src)) and (dst in [int, bool]):
        # Pure numeric or enum can be cast to a numeric
        return True
    elif isinstance(src, rdltypes.ArrayPlaceholder) and isinstance(dst, rdltypes.ArrayPlaceholder):
        # Check that array element types also match
        if src.element_type is None:
            # indeterminate array type. Is castable
            return True
        return is_castable(src.element_type, dst.element_type)
    elif rdltypes.is_user_struct(dst):
        # Structs can be assigned their derived counterparts - aka their subclasses
        return issubclass(src, dst)
    elif dst == rdltypes.PropertyReference:
        return issubclass(src, rdltypes.PropertyReference)
    elif dst == rdltypes.references.RefType:
        # Any reference
        if issubclass(src, comp.Component):
            return True
        return issubclass(src, rdltypes.PropertyReference)
    elif src == dst:
        return True
    else:
        return False
